// ********************************************************************
// *                                                                  *
// * Rauch zwischen zwei Glasplatten von TOBIAS HERMANN nach Jos Stam *
// *                                                                  *
// ********************************************************************
// *                                                                  *
// *     Steuerung: w,a,s,d + pfeiltasten    fr die Kamera           *
// *     c = mond zurcksetzen und nach links drehen                  *
// *     t = mond zurcksetzen und nach rechts drehen                 *
// *     "+" und "-" = simulationsgeschwindigkeit ndern              *
// *                                                                  *
// ********************************************************************

#ifdef _MSC_VER
	#pragma comment( linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"" )
	#pragma comment( lib, "opengl32.lib" )
	#pragma comment( lib, "glu32.lib" )
	#pragma comment( lib, "glut32.lib" )
	#pragma comment( lib, "glaux.lib" )
#endif

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "CWin32Mouse.h"
#include "CCameraControl.h"
#include "tgaload.h"

#define GAMEMODE_STRING	"1024x768:32"

void initEnvironment(void);
void renderScene(void);
void drawOrigin(void);
void changePerspective(int width, int height);
void normalKeyPressed(unsigned char key, int x, int y);
void normalKeyReleased(unsigned char key, int x, int y);
void specialKeyPressed(int key, int x, int y);
void specialKeyReleased(int key, int x, int y);

CCameraControl cameraControl;
CWin32Mouse    mouse;
POINT          mP;
int mouseActive = 0;
int isGridActive = 0;
float rotateAngle = 0.0f;
image_t earth_image, moon_image, space_image;

int einatmen = 0;
bool schornstein_an = false;

bool moon_am_aufladen = false;
float moon_auflade_richtung = 1;


#define win_width 800
#define win_height 600





#define IX(i,j) ((i)+(N+2)*(j))


static int N;
static float dt, diff, visc;
static float force, source;
static int dvel;

static float * u, * v, * u_prev, * v_prev;
static float * dens, * dens_prev;

static int win_id;
static int win_x, win_y;
static int mouse_down[3];
static int omx, omy, mx, my;


#define SWAP(x0,x) {float * tmp=x0;x0=x;x=tmp;}
#define FOR_EACH_CELL for ( i=1 ; i<=N ; i++ ) { for ( j=1 ; j<=N ; j++ ) {
#define END_FOR }}

void add_source ( int N, float * x, float * s, float dt )
{
	int i, size=(N+2)*(N+2);
	for ( i=0 ; i<size ; i++ ) x[i] += dt*s[i];
}

void set_bnd ( int N, int b, float * x )
{
	int i;

	for ( i=1 ; i<=N ; i++ ) {
		x[IX(0  ,i)] = b==1 ? -x[IX(1,i)] : x[IX(1,i)];
		x[IX(N+1,i)] = b==1 ? -x[IX(N,i)] : x[IX(N,i)];
		x[IX(i,0  )] = b==2 ? -x[IX(i,1)] : x[IX(i,1)];
		x[IX(i,N+1)] = b==2 ? -x[IX(i,N)] : x[IX(i,N)];
	}
	x[IX(0  ,0  )] = 0.5f*(x[IX(1,0  )]+x[IX(0  ,1)]);
	x[IX(0  ,N+1)] = 0.5f*(x[IX(1,N+1)]+x[IX(0  ,N)]);
	x[IX(N+1,0  )] = 0.5f*(x[IX(N,0  )]+x[IX(N+1,1)]);
	x[IX(N+1,N+1)] = 0.5f*(x[IX(N,N+1)]+x[IX(N+1,N)]);
}

void lin_solve ( int N, int b, float * x, float * x0, float a, float c )
{
	int i, j, k;

	for ( k=0 ; k<20 ; k++ ) {
		FOR_EACH_CELL
			x[IX(i,j)] = (x0[IX(i,j)] + a*(x[IX(i-1,j)]+x[IX(i+1,j)]+x[IX(i,j-1)]+x[IX(i,j+1)]))/c;
		END_FOR
		set_bnd ( N, b, x );
	}
}

void diffuse ( int N, int b, float * x, float * x0, float diff, float dt )
{
	float a=dt*diff*N*N;
	lin_solve ( N, b, x, x0, a, 1+4*a );
}

void advect ( int N, int b, float * d, float * d0, float * u, float * v, float dt )
{
	int i, j, i0, j0, i1, j1;
	float x, y, s0, t0, s1, t1, dt0;

	dt0 = dt*N;
	FOR_EACH_CELL
		x = i-dt0*u[IX(i,j)]; y = j-dt0*v[IX(i,j)];
		if (x<0.5f) x=0.5f; if (x>N+0.5f) x=N+0.5f; i0=(int)x; i1=i0+1;
		if (y<0.5f) y=0.5f; if (y>N+0.5f) y=N+0.5f; j0=(int)y; j1=j0+1;
		s1 = x-i0; s0 = 1-s1; t1 = y-j0; t0 = 1-t1;
		d[IX(i,j)] = s0*(t0*d0[IX(i0,j0)]+t1*d0[IX(i0,j1)])+
					 s1*(t0*d0[IX(i1,j0)]+t1*d0[IX(i1,j1)]);
	END_FOR
	set_bnd ( N, b, d );
}

void project ( int N, float * u, float * v, float * p, float * div )
{
	int i, j;

	FOR_EACH_CELL
		div[IX(i,j)] = -0.5f*(u[IX(i+1,j)]-u[IX(i-1,j)]+v[IX(i,j+1)]-v[IX(i,j-1)])/N;
		p[IX(i,j)] = 0;
	END_FOR	
	set_bnd ( N, 0, div ); set_bnd ( N, 0, p );

	lin_solve ( N, 0, p, div, 1, 4 );

	FOR_EACH_CELL
		u[IX(i,j)] -= 0.5f*N*(p[IX(i+1,j)]-p[IX(i-1,j)]);
		v[IX(i,j)] -= 0.5f*N*(p[IX(i,j+1)]-p[IX(i,j-1)]);
	END_FOR
	set_bnd ( N, 1, u ); set_bnd ( N, 2, v );
}

void dens_step ( int N, float * x, float * x0, float * u, float * v, float diff, float dt )
{
	add_source ( N, x, x0, dt );
	SWAP ( x0, x ); diffuse ( N, 0, x, x0, diff, dt );
	SWAP ( x0, x ); advect ( N, 0, x, x0, u, v, dt );
}

void vel_step ( int N, float * u, float * v, float * u0, float * v0, float visc, float dt )
{
	add_source ( N, u, u0, dt ); add_source ( N, v, v0, dt );
	SWAP ( u0, u ); diffuse ( N, 1, u, u0, visc, dt );
	SWAP ( v0, v ); diffuse ( N, 2, v, v0, visc, dt );
	project ( N, u, v, u0, v0 );
	SWAP ( u0, u ); SWAP ( v0, v );
	advect ( N, 1, u, u0, u0, v0, dt ); advect ( N, 2, v, v0, u0, v0, dt );
	project ( N, u, v, u0, v0 );
}



static void free_data ( void )
{
	if ( u ) free ( u );
	if ( v ) free ( v );
	if ( u_prev ) free ( u_prev );
	if ( v_prev ) free ( v_prev );
	if ( dens ) free ( dens );
	if ( dens_prev ) free ( dens_prev );
}

static void clear_data ( void )
{
	int i, size=(N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = u_prev[i] = v_prev[i] = dens[i] = dens_prev[i] = 0.0f;
	}
}

static int allocate_data ( void )
{
	int size = (N+2)*(N+2);

	u			= (float *) malloc ( size*sizeof(float) );
	v			= (float *) malloc ( size*sizeof(float) );
	u_prev		= (float *) malloc ( size*sizeof(float) );
	v_prev		= (float *) malloc ( size*sizeof(float) );
	dens		= (float *) malloc ( size*sizeof(float) );	
	dens_prev	= (float *) malloc ( size*sizeof(float) );

	if ( !u || !v || !u_prev || !v_prev || !dens || !dens_prev ) {
		fprintf ( stderr, "cannot allocate data\n" );
		return ( 0 );
	}

	return ( 1 );
}


/*
  ----------------------------------------------------------------------
   OpenGL specific drawing routines
  ----------------------------------------------------------------------
*/

static void pre_display ( void )
{
	glViewport ( 0, 0, win_x, win_y );
	glMatrixMode ( GL_PROJECTION );
	glLoadIdentity ();
	gluOrtho2D ( 0.0, 1.0, 0.0, 1.0 );
	glClearColor ( 0.0f, 0.0f, 0.0f, 1.0f );
	glClear ( GL_COLOR_BUFFER_BIT );
}

static void post_display ( void )
{
	glutSwapBuffers ();
}

static void draw_velocity ( void )
{
	int i, j;
	float x, y, h;

	h = 1.0f/N;

	glColor4f ( 1.0f, 0.0f, 0.0f, 0.4f );
	glLineWidth ( 1.0f );

	glBegin ( GL_LINES );

		for ( i=1 ; i<=N ; i++ ) {
			x = (i-0.5f)*h;
			for ( j=1 ; j<=N ; j++ ) {
				y = (j-0.5f)*h;

				glVertex3f ( x, y, 0 );
				glVertex3f ( x+u[IX(i,j)], y+v[IX(i,j)], 0 );
			}
		}

	glEnd ();
}


static void draw_density ( void )
{
	int i, j;
	float x, y, h, d00, d01, d10, d11;

	h = 1.0f/N;

	glBegin ( GL_QUADS );

		for ( i=0 ; i<=N ; i++ ) {
			x = (i-0.5f)*h;
			for ( j=0 ; j<=N ; j++ ) {
				y = (j-0.5f)*h;

				d00 = dens[IX(i,j)];
				d01 = dens[IX(i,j+1)];
				d10 = dens[IX(i+1,j)];
				d11 = dens[IX(i+1,j+1)];

				glColor4f ( 1.0f, 1.0f, 1.0f, d00 ); glVertex3f ( x,   y,  0 );
				glColor4f ( 1.0f, 1.0f, 1.0f, d10 ); glVertex3f ( x+h, y,  0 );
				glColor4f ( 1.0f, 1.0f, 1.0f, d11 ); glVertex3f ( x+h, y+h,0 );
				glColor4f ( 1.0f, 1.0f, 1.0f, d01 ); glVertex3f ( x,   y+h,0 );
			}
		}

	glEnd ();
}

/*
  ----------------------------------------------------------------------
   relates mouse movements to forces sources
  ----------------------------------------------------------------------
*/

static void get_from_UI ( float * d, float * u, float * v )
{
	int i, j, size = (N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] = 0.0f;
	}

	if ( !mouse_down[0] && !mouse_down[2] ) return;

	i = (int)((       mx /(float)win_x)*N+1);
	j = (int)(((win_y-my)/(float)win_y)*N+1);

	if ( i<1 || i>N || j<1 || j>N ) return;

	if ( mouse_down[0] ) {
		u[IX(i,j)] = force * (mx-omx);
		v[IX(i,j)] = force * (omy-my);
	}

	if ( mouse_down[2] ) {
		d[IX(i,j)] = source;
	}

	omx = mx;
	omy = my;

	return;
}

/*
  ----------------------------------------------------------------------
   GLUT callback routines
  ----------------------------------------------------------------------
*/


static void mouse_func ( int button, int state, int x, int y )
{
	omx = mx = x;
	omx = my = y;

	mouse_down[button] = state == GLUT_DOWN;
}

static void motion_func ( int x, int y )
{
	mx = x;
	my = y;
}

static void reshape_func ( int width, int height )
{
	glutSetWindow ( win_id );
	glutReshapeWindow ( width, height );

	win_x = width;
	win_y = height;
}

static void idle_func ( void )
{
	get_from_UI ( dens_prev, u_prev, v_prev );
	vel_step ( N, u, v, u_prev, v_prev, visc, dt );
	dens_step ( N, dens, dens_prev, u, v, diff, dt );

	glutSetWindow ( win_id );
	glutPostRedisplay ();
}

static void display_func ( void )
{
	pre_display ();

		if ( dvel ) draw_velocity ();
		else		draw_density ();

	post_display ();
}


float zufallszahl( float ende )
{
	float wert;
	wert = ( rand() * ende ) / 32767;
	return wert;
}

static void mache_durcheinander ( float * d, float * u, float * v, float staerke )
{
	int i, j, size = (N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] = 0.0f;
	}


	i = (int)zufallszahl(N-1)+1;
	j = (int)zufallszahl(N-1+1);

	if ( i<1 || i>N || j<1 || j>N ) return;

	
		u[IX(i,j)] = staerke * force * (zufallszahl(10)-5);
		v[IX(i,j)] = staerke * force * (zufallszahl(10)-5);
		
		d[IX(i,j)] = source * staerke;
		


	return;
}

static void schornstein ( float * d, float * u, float * v, float staerke )
{
	int i, j, size = (N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] = 0.0f;
	}


	i = (int)zufallszahl(4)+2+N/2;
	j = 2;

	if ( i<1 || i>N || j<1 || j>N ) return;

	
		//u[IX(i,j)] = staerke * force * (zufallszahl(10)-5);
	u[IX(i,j)] = 0;
	v[IX(i,j)] = staerke * force * 2;
		
		d[IX(i,j)] = source * staerke / 5;
		
	
	return;
}



static void mache_ordnung ( float * d, float * u, float * v )
{
	int i, j, size = (N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] = 0.0f;
	}
	return;
}

void output(float x, float y, char *string)
{
	int len, i;
	glRasterPos3f(x, y, -1);
	len = (int) strlen(string);
	for (i = 0; i < len; i++) {
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, string[i]);
	}
}


void renderScene(void)
{
	char str[255];
	if (mouseActive==1)
	{
		POINT val = mouse.getRelativeMovement();
		cameraControl.mouseTurn(val.x);
		cameraControl.mouseTilt(-val.y);
	}
	cameraControl.checkCurrentMovement();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();
		glLoadIdentity();
		glColor3f(0.6f, 0.7f, 1.0f);
		sprintf (str,"keys: T,V, w,a,s,d,pfeiltasten");
		output(-0.49f,-0.4f,str);
	glPopMatrix();
	


	
		if ( dvel ) draw_velocity ();
		else		draw_density ();

if (schornstein_an)  schornstein(dens_prev, u_prev, v_prev, 1);

	vel_step ( N, u, v, u_prev, v_prev, visc, dt );
	dens_step ( N, dens, dens_prev, u, v, diff, dt );
	mache_ordnung ( dens_prev, u_prev, v_prev );

	
	glutSwapBuffers();
}






void main(int argc, char *argv[])
{
	

	

	N = 64;
		dt = 0.1f;
		diff = 0.0f;
		visc = 0.0f;
		force = 5.0f;
		source = 100.0f;

		if ( !allocate_data () )	exit ( 1 );
		clear_data ();


		srand( (unsigned)time( NULL ) );


	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowPosition(100, 100);
	glutInitWindowSize(win_width, win_height);	
	glutCreateWindow("Tobias Hermann");

	tgaGetColorEXT();
	glColorMaterial ( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );

		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		glEnable(GL_BLEND);
		glDisable(GL_DEPTH_TEST);

	glEnable      ( GL_TEXTURE_2D );
	glPixelStorei ( GL_UNPACK_ALIGNMENT, 1 );
	glGenTextures ( 3, 0 );

	cameraControl.eyePosZ = 1.4;
	cameraControl.eyePosX = 0.5;
	cameraControl.eyePosY = 0.5;
		
	initEnvironment();

	glutMainLoop();

	free_data ();
}

void initEnvironment(void)
{
	glutDisplayFunc(renderScene);
	glutIdleFunc(renderScene);
	glutReshapeFunc(changePerspective);
	glutIgnoreKeyRepeat(1);
	glutKeyboardFunc(normalKeyPressed);
	glutKeyboardUpFunc(normalKeyReleased);
	glutSpecialFunc(specialKeyPressed);
	glutSpecialUpFunc(specialKeyReleased);
	//glEnable(GL_TEXTURE_2D);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
}



void changePerspective(int width, int height)
{
	if (height == 0) height = 1;
	float ratio = 1.0f * width / height;
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0, 0, width, height);
	gluPerspective(45.0f, ratio, 0.1f, 1000.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	cameraControl.updateCamera();
}


void normalKeyPressed(unsigned char key, int x, int y)
{
	switch (key)
	{
		//Programm mit ESC beenden
		case 27:
			free_data ();
			exit(0);
			break;
		// KAMERA Steuerung
		case 'w':
			cameraControl.walkForward();
			break;
		
		case 's':
			cameraControl.walkBackward();
			break;
			
		case 'a':
			cameraControl.strafeLeft();
			break;
		
		case 'd':
			cameraControl.strafeRight();
			break;

		case 'e':
			einatmen ++;			
			break;

		case 'r':
			mache_durcheinander ( dens_prev, u_prev, v_prev, einatmen );
			einatmen = 0;
			break;

		case 't':
			schornstein_an = !schornstein_an;
			break;


		case 'c':
		case 'C':
			clear_data ();
			break;
	
		case 'v':
		case 'V':
			dvel = !dvel;
			break;
	}
}

void normalKeyReleased(unsigned char key, int x, int y)
{
	switch (key)
	{
		// KAMERA Steuerung
		case 'w':
		case 'W':
		case 's':
		case 'S':
			cameraControl.stopWalking();
			cameraControl.stopFlying();
			break;
		
		case 'a':
		case 'A':
		case 'd':
		case 'D':
			cameraControl.stopStrafing();
			break;
		case 'r':
			moon_am_aufladen = false;
			break;
		case 't':
			moon_am_aufladen = false;
			break;
		case 'g':
			if (isGridActive)
				isGridActive = 0;
			else
				isGridActive = 1;
			break;

		case ' ':
			if (mouseActive == 1)
			{
				glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
				mouseActive = 0;
			}
			else
			{
				glutSetCursor(GLUT_CURSOR_NONE);
				mouseActive = 1;
			}
			break;	
	}
}

void specialKeyPressed(int key, int x, int y)
{
	int modifiers = glutGetModifiers();

	switch (key)
	{
		// Zwischen Fullscreen und Window wechseln
		case GLUT_KEY_F11:
			if (glutGameModeGet(GLUT_GAME_MODE_ACTIVE))
			{
				glutLeaveGameMode();
			}
			else
			{
				glutGameModeString(GAMEMODE_STRING);
				if (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE))
				{
					glutEnterGameMode();

					if (mouseActive == 1)
						glutSetCursor(GLUT_CURSOR_NONE);
					else
						glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
										
					mouse.resetMouse();
					initEnvironment();
				}
				else
					exit(1);
			}
			break;
		
		// KAMERA Steuerung
		case GLUT_KEY_LEFT:
			cameraControl.turnLeft();
			break;
		
		case GLUT_KEY_RIGHT:
			cameraControl.turnRight();
			break;
		
		case GLUT_KEY_UP:
			cameraControl.tiltUp();
			break;
		
		case GLUT_KEY_DOWN:
			cameraControl.tiltDown();
			break;

		case GLUT_KEY_PAGE_UP:
			cameraControl.hoverUp();
			break;
		
		case GLUT_KEY_PAGE_DOWN:
			cameraControl.hoverDown();
			break;
	}
}

void specialKeyReleased(int key, int x, int y)
{
	switch (key)
	{
		// KAMERA Steuerung
		case GLUT_KEY_LEFT:
		case GLUT_KEY_RIGHT:
			cameraControl.stopTurning();
			break;
		
		case GLUT_KEY_UP:
		case GLUT_KEY_DOWN:
			cameraControl.stopTilting();
			break;
		
		case GLUT_KEY_PAGE_UP:
		case GLUT_KEY_PAGE_DOWN:
			cameraControl.stopHovering();
			break;
	}
}