//*****************************************************************************************
//*****************************************************************************************
//*                   CCameraControl 1.2 (C) Blackwidow and Finalbrain                    *
//*****************************************************************************************
//*****************************************************************************************

#include "CCameraControl.h"
#include <GL/glut.h>
#include <math.h>
#include <stdio.h>


// **********************************************************
// ****************** FFENTLICHE METHODEN ******************
// **********************************************************

CCameraControl::CCameraControl()
{
	initPerspective();
}

void CCameraControl::updateCamera()
{
	gluLookAt(eyePosX, eyePosY, eyePosZ,
		      eyePosX + lineOfSightX, eyePosY + lineOfSightY, eyePosZ + lineOfSightZ,
			  0.0f, 1.0f, 0.0f);
}

void CCameraControl::checkCurrentMovement()
{
	if (viewIsTurning)
		turnView(viewIsTurning);
	
	if (viewIsTilting)
		tiltView(viewIsTilting);
	
	if (isWalking)
		walk(isWalking);
	
	if (isFlying)
		fly(isFlying);
	
	if (isStrafing)
		strafe(isStrafing);

	if (isHovering)
		hover(isHovering);
}

void CCameraControl::turnLeft()
{
	if (viewIsTurning == TURN_RIGHT)
		turningCache = TURN_LEFT;
	else
		viewIsTurning = TURN_LEFT;
}

void CCameraControl::turnRight()
{
	if (viewIsTurning == TURN_LEFT)
		turningCache = TURN_RIGHT;
	else
		viewIsTurning = TURN_RIGHT;
}

void CCameraControl::stopTurning()
{
	if (turningCache == STOP_TURNING)
		viewIsTurning = STOP_TURNING;
	else
	{
		viewIsTurning = turningCache;
		turningCache = STOP_TURNING;
	}
}

void CCameraControl::tiltUp()
{
	if (viewIsTilting == TILT_DOWN)
		tiltingCache = TILT_UP;
	else
		viewIsTilting = TILT_UP;
}

void CCameraControl::tiltDown()
{
	if (viewIsTilting == TILT_UP)
		tiltingCache = TILT_DOWN;
	else
		viewIsTilting = TILT_DOWN;
}

void CCameraControl::stopTilting()
{
	if (tiltingCache == STOP_TILTING)
		viewIsTilting = STOP_TILTING;
	else
	{
		viewIsTilting = tiltingCache;
		tiltingCache = STOP_TILTING;
	}
}

void CCameraControl::walkForward()
{
	if (isWalking == WALK_BACKWARD)
		walkingCache = WALK_FORWARD;
	else
		isWalking = WALK_FORWARD;
}

void CCameraControl::walkBackward()
{
	if (isWalking == WALK_FORWARD)
		walkingCache = WALK_BACKWARD;
	else
		isWalking = WALK_BACKWARD;
}

void CCameraControl::stopWalking()
{
	if (walkingCache == STOP_WALKING)
		isWalking = STOP_WALKING;
	else
	{
		isWalking = walkingCache;
		walkingCache = STOP_WALKING;
	}
}

void CCameraControl::flyForward()
{
	if (isFlying == FLY_BACKWARD)
		flyingCache = FLY_FORWARD;
	else
		isFlying = FLY_FORWARD;
}

void CCameraControl::flyBackward()
{
	if (isFlying == FLY_FORWARD)
		flyingCache = FLY_BACKWARD;
	else
		isFlying = FLY_BACKWARD;
}

void CCameraControl::stopFlying()
{
	if (flyingCache == STOP_FLYING)
		isFlying = STOP_FLYING;
	else
	{
		isFlying = flyingCache;
		flyingCache = STOP_FLYING;
	}
}

void CCameraControl::strafeLeft()
{
	if (isStrafing == STRAFE_RIGHT)
		strafingCache = STRAFE_LEFT;
	else
		isStrafing = STRAFE_LEFT;
}

void CCameraControl::strafeRight()
{
	if (isStrafing == STRAFE_LEFT)
		strafingCache = STRAFE_RIGHT;
	else
		isStrafing = STRAFE_RIGHT;
}

void CCameraControl::stopStrafing()
{
	if (strafingCache == STOP_STRAFING)
		isStrafing = STOP_STRAFING;
	else
	{
		isStrafing = strafingCache;
		strafingCache = STOP_STRAFING;
	}
}

void CCameraControl::hoverUp()
{
	if (isHovering == HOVER_DOWN)
		hoveringCache = HOVER_UP;
	else
		isHovering = HOVER_UP;
}

void CCameraControl::hoverDown()
{
	if (isHovering == HOVER_UP)
		hoveringCache = HOVER_DOWN;
	else
		isHovering = HOVER_DOWN;
}

void CCameraControl::stopHovering()
{
	if (hoveringCache == STOP_HOVERING)
		isHovering = STOP_HOVERING;
	else
	{
		isHovering = hoveringCache;
		hoveringCache = STOP_HOVERING;
	}
}

void CCameraControl::mouseTurn(int pixels)
{
	int RATIOTURN = 1;
	int STEPSTURN = pixels / RATIOTURN;
	turnView(STEPSTURN);
}

void CCameraControl::mouseTilt(int pixels)
{
	int RATIOTILT = 1;
	int STEPSTILT = pixels / RATIOTILT;
	tiltView(STEPSTILT);
}


// **********************************************************
// ******************** PRIVATE METHODEN ********************
// **********************************************************

void CCameraControl::initPerspective()
{
	eyePosX = 0.0f;
	eyePosY = 0.0f;
	eyePosZ = 2.0f;

	turnViewAngle = 0;
	tiltViewAngle = 0;
	
	viewIsTurning = STOP_TURNING;
	turningCache = STOP_TURNING;
	viewIsTilting = STOP_TILTING;
	tiltingCache = STOP_TILTING;
	isWalking = STOP_WALKING;
	walkingCache = STOP_WALKING;
	isFlying = STOP_FLYING;
	flyingCache = STOP_FLYING;
	isStrafing = STOP_STRAFING;
	strafingCache = STOP_STRAFING;
	isHovering = STOP_HOVERING;
	hoveringCache = STOP_HOVERING;
	
	lineOfSightX = sin(turnViewAngle);
	lineOfSightY = 0.0f;
	lineOfSightZ = -cos(turnViewAngle);
}

void CCameraControl::turnView(int steps)
{
	turnViewAngle += (TURNING_STEP * steps);
	lineOfSightX = sin(turnViewAngle);
	lineOfSightZ = -cos(turnViewAngle);
	glLoadIdentity();
	updateCamera();
}

void CCameraControl::tiltView(int steps)
{
	tiltViewAngle += (TILTING_STEP * steps);
	
	if (tiltViewAngle > TILTING_MAX)
	{
		tiltViewAngle = TILTING_MAX;
		viewIsTilting = STOP_TILTING;
	}
	else if (tiltViewAngle < TILTING_MIN)
	{
		tiltViewAngle = TILTING_MIN;
		viewIsTilting = STOP_TILTING;
	}
	
	lineOfSightY = sin(tiltViewAngle);
	glLoadIdentity();
	updateCamera();
}

void CCameraControl::walk(int steps)
{
	eyePosX = eyePosX + steps * (lineOfSightX) * WALKING_STEP;
	eyePosZ = eyePosZ + steps * (lineOfSightZ) * WALKING_STEP;
	glLoadIdentity();
	updateCamera();
}

void CCameraControl::fly(int steps)
{
	eyePosX = eyePosX + steps * (lineOfSightX) * FLYING_STEP;
	eyePosY = eyePosY + steps * (lineOfSightY) * FLYING_STEP;
	eyePosZ = eyePosZ + steps * (lineOfSightZ) * FLYING_STEP;
	glLoadIdentity();
	updateCamera();
}

void CCameraControl::strafe(int steps)
{
	eyePosX = eyePosX + steps * (lineOfSightZ) * STRAFING_STEP;
	eyePosZ = eyePosZ + (-steps) * (lineOfSightX) * STRAFING_STEP;
	glLoadIdentity();
	updateCamera();
}

void CCameraControl::hover(int steps)
{
	eyePosY = eyePosY + steps * HOVERING_STEP;
	glLoadIdentity();
	updateCamera();
}
