// ********************************************************************
// *                                                                  *
// * Rauch 3d von TOBIAS HERMANN nach Jos Stam                        *
// *                                                                  *
// ********************************************************************
// *                                                                  *
// *     Steuerung: w,a,s,d + pfeiltasten    fr die Kamera           *
// *     h = darstellungsart ndern                                   *
// *     t = schornstein ein und ausschalten                          *
// *     "+" und "-" = N verndern                                    *
// *                                                                  *
// ********************************************************************

#ifdef _MSC_VER
	#pragma comment( linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"" )
	#pragma comment( lib, "opengl32.lib" )
	#pragma comment( lib, "glu32.lib" )
	#pragma comment( lib, "glut32.lib" )
	#pragma comment( lib, "glaux.lib" )
#endif

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <math.h>
#include "CWin32Mouse.h"
#include "CCameraControl.h"
#include "tgaload.h"


#define GAMEMODE_STRING	"1024x768:32"

void initEnvironment(void);
void renderScene(void);
void drawOrigin(void);
void changePerspective(int width, int height);
void normalKeyPressed(unsigned char key, int x, int y);
void normalKeyReleased(unsigned char key, int x, int y);
void specialKeyPressed(int key, int x, int y);
void specialKeyReleased(int key, int x, int y);

CCameraControl cameraControl;
CWin32Mouse    mouse;
POINT          mP;
int mouseActive = 0;
int isGridActive = 0;
float rotateAngle = 0.0f;
image_t earth_image, moon_image, space_image;

int einatmen = 0;
bool schornstein_an = false;
int fps = 0;
int framecounter = 0;
int sekunden = 0;
int sekunden_alt = 0;

#define win_width 800
#define win_height 600



#define IX3(i,j,k) ( (i) + ( (N+2)*(j) ) + ( (N+2)*(N+2)*(k) ) )
#define IX(i,j) (IX3(i,j,N/2))


static int N = 18;
static int z_stretch = 1;
static float dt, diff, visc;
static float force, source;
static int * object;
static int dvel;
static int draw_dens_mode = 0;

static float * u, * v, * w, * u_prev, * v_prev, * w_prev;
static float * dens, * dens_prev;

static int win_id;
static int win_x, win_y;
static int mouse_down[3];
static int omx, omy, mx, my;
int obj_x, obj_y, obj_z, obj_s;


#define SWAP(x0,x) {float * tmp=x0;x0=x;x=tmp;}
#define FOR_EACH_CELL for ( i=1 ; i<=N ; i++ ) { for ( j=1 ; j<=N ; j++ ) {
#define FOR_EACH_CELL3 for ( i=1 ; i<=N ; i++ ) { for ( j=1 ; j<=N ; j++ ) { for ( k=1 ; k<=N ; k++ ) {
#define END_FOR }}
#define END_FOR3 }}}



void add_source3 ( int N, float * x, float * s, float dt )
{
	int i, size=(N+2)*(N+2)*(N+2);
	for ( i=0 ; i<size ; i++ ) x[i] += dt*s[i];
}

void set_bnd_objekt3 ( int N, int b, float * x ) // b == 1 bei u ; b == 2 bei v
{
	int i,j,k;

	for ( i=1 ; i<=N ; i++ ) {
		for ( j=1 ; j<=N ; j++ ) {
			for ( k=1 ; k<=N ; k++ ) {
				if ( object[IX(i,j)] > 0 ) {
					x[IX3(i,j,k)] = b>0 ? -x[IX3(i,j,k)] : x[IX3(i,j,k)];
				}
			}
		}
	}
}

void set_bnd3 ( int N, int b, float * x )
{
	int i,j;

	for ( i=1 ; i<=N ; i++ ) {
		for ( j=1 ; j<=N ; j++ ) {
			//x[IX3(0  ,i,j)] = b==1 ? -x[IX3(1,i,j)] : x[IX3(1,i,j)];
			x[IX3(0  ,i,j)] = b==1 ? 0.5f : x[IX3(1,i,j)];
			x[IX3(N+1,i,j)] = b==1 ? -x[IX3(N,i,j)] : x[IX3(N,i,j)];
			x[IX3(i,0  ,j)] = b==2 ? -x[IX3(i,1,j)] : x[IX3(i,1,j)];
			x[IX3(i,N+1,j)] = b==2 ? -x[IX3(i,N,j)] : x[IX3(i,N,j)];
			x[IX3(i,j,0  )] = b==3 ? -x[IX3(i,j,1)] : x[IX3(i,j,1)];
			x[IX3(i,j,N+1)] = b==3 ? -x[IX3(i,j,N)] : x[IX3(i,j,N)];
		}
	}

	for ( i=1 ; i<=N ; i++ ) {
		x[IX3(0  ,0  ,i)] = 0.5f*(x[IX3(1,0  ,i)]+x[IX3(0  ,1,i)]);
		x[IX3(0  ,N+1,i)] = 0.5f*(x[IX3(1,N+1,i)]+x[IX3(0  ,N,i)]);
		x[IX3(N+1,0  ,i)] = 0.5f*(x[IX3(N,0  ,i)]+x[IX3(N+1,1,i)]);
		x[IX3(N+1,N+1,i)] = 0.5f*(x[IX3(N,N+1,i)]+x[IX3(N+1,N,i)]);

		x[IX3(i, 0  ,0  )] = 0.5f*(x[IX3(i, 1,0  )]+x[IX3(i, 0  ,1)]);
		x[IX3(i, 0  ,N+1)] = 0.5f*(x[IX3(i, 1,N+1)]+x[IX3(i, 0  ,N)]);
		x[IX3(i, N+1,0  )] = 0.5f*(x[IX3(i, N,0  )]+x[IX3(i, N+1,1)]);
		x[IX3(i, N+1,N+1)] = 0.5f*(x[IX3(i, N,N+1)]+x[IX3(i, N+1,N)]);

		x[IX3(0  ,i,0  )] = 0.5f*(x[IX3(1,i,0  )]+x[IX3(0  ,i,1)]);
		x[IX3(0  ,i,N+1)] = 0.5f*(x[IX3(1,i,N+1)]+x[IX3(0  ,i,N)]);
		x[IX3(N+1,i,0  )] = 0.5f*(x[IX3(N,i,0  )]+x[IX3(N+1,i,1)]);
		x[IX3(N+1,i,N+1)] = 0.5f*(x[IX3(N,i,N+1)]+x[IX3(N+1,i,N)]);
	}

	x[IX3(0  ,0  ,0  )] = (1/3)*(x[IX3(1  ,0  ,0  )]+x[IX3(0  ,1  ,0  )]+x[IX3(0  ,0  ,1  )]);
	x[IX3(N+1,0  ,0  )] = (1/3)*(x[IX3(N  ,0  ,0  )]+x[IX3(N+1,1  ,0  )]+x[IX3(N+1,0  ,1  )]);
	x[IX3(0  ,N+1,0  )] = (1/3)*(x[IX3(1  ,N+1,0  )]+x[IX3(0  ,N  ,0  )]+x[IX3(0  ,0  ,N+1)]);
	x[IX3(0  ,0  ,N+1)] = (1/3)*(x[IX3(1  ,0  ,N+1)]+x[IX3(0  ,N  ,N+1)]+x[IX3(0  ,0  ,N  )]);
	x[IX3(N+1,0  ,N+1)] = (1/3)*(x[IX3(N  ,0  ,N+1)]+x[IX3(N+1,N  ,N+1)]+x[IX3(N+1,0  ,N  )]);
	x[IX3(N+1,N+1,0  )] = (1/3)*(x[IX3(N  ,N+1,0  )]+x[IX3(N+1,N  ,0  )]+x[IX3(N+1,N+1,1  )]);
	x[IX3(0  ,N+1,N+1)] = (1/3)*(x[IX3(1  ,N+1,N+1)]+x[IX3(0  ,N  ,N+1)]+x[IX3(0  ,N+1,N  )]);
	x[IX3(N+1,N+1,N+1)] = (1/3)*(x[IX3(N  ,N+1,N+1)]+x[IX3(N+1,N  ,N+1)]+x[IX3(N+1,N+1,N  )]);

	set_bnd_objekt3(N,b,x);
}





void lin_solve3 ( int N, int b, float * x, float * x0, float a, float c )
{
	int i, j, k, k1;

	for ( k1=0 ; k1<20 ; k1++ ) {
		FOR_EACH_CELL3
			x[IX3(i,j,k)] = (x0[IX3(i,j,k)] + a*(x[IX3(i-1,j,k)]+x[IX3(i+1,j,k)]+x[IX3(i,j-1,k)]+x[IX3(i,j+1,k)]+x[IX3(i,j,k-1)]+x[IX3(i,j,k+1)]))/c;			
		END_FOR3
		set_bnd3 ( N, b, x );
	}
}



void diffuse3 ( int N, int b, float * x, float * x0, float diff, float dt )
{
	float a=dt*diff*N*N*N;
	lin_solve3 ( N, b, x, x0, a, 1+6*a );
}




void advect3 ( int N, int b, float * d, float * d0, float * u, float * v, float * w, float dt )
{
	int i, j, k, i0, j0, k0, i1, j1, k1;
	float x, y, z, s0, t0, r0, s1, t1, r1, dt0;

	dt0 = dt*N;
	FOR_EACH_CELL3
		x = i-dt0*u[IX3(i,j,k)]; y = j-dt0*v[IX3(i,j,k)]; z = k-dt0*w[IX3(i,j,k)];
		if (x<0.5f) x=0.5f; if (x>N+0.5f) x=N+0.5f; i0=(int)x; i1=i0+1;
		if (y<0.5f) y=0.5f; if (y>N+0.5f) y=N+0.5f; j0=(int)y; j1=j0+1;
		if (z<0.5f) z=0.5f; if (z>N+0.5f) z=N+0.5f; k0=(int)z; k1=k0+1;
		s1 = x-i0; s0 = 1-s1; t1 = y-j0; t0 = 1-t1; r1 = z-k0; r0 = 1-r1;	
		d[IX3(i,j,k)] = s0 * t0 * r0 * d0[IX3(i0,j0,k0)] +
						s1 * t0 * r0 * d0[IX3(i1,j0,k0)] +
						s0 * t1 * r0 * d0[IX3(i0,j1,k0)] +
						s0 * t0 * r1 * d0[IX3(i0,j0,k1)] +
						s1 * t1 * r0 * d0[IX3(i1,j1,k0)] +
						s0 * t1 * r1 * d0[IX3(i0,j1,k1)] +
						s1 * t0 * r1 * d0[IX3(i1,j0,k1)] +
						s1 * t1 * r1 * d0[IX3(i1,j1,k1)];
	
	END_FOR3
	set_bnd3 ( N, b, d );
}


void project3 ( int N, float * u, float * v, float * w, float * p, float * div )
{
	int i, j, k;

	FOR_EACH_CELL3
		div[IX3(i,j,k)] = -(1.0f/2.0f)*(u[IX3(i+1,j,k)]-u[IX3(i-1,j,k)]+v[IX3(i,j+1,k)]-v[IX3(i,j-1,k)]+w[IX3(i,j,k+1)]-w[IX3(i,j,k-1)])/N;
		p[IX3(i,j,k)] = 0;
	END_FOR3
	set_bnd3 ( N, 0, div ); set_bnd3 ( N, 0, p );

	lin_solve3 ( N, 0, p, div, 1, 6 );
	FOR_EACH_CELL3
		u[IX3(i,j,k)] -= (1.0f/2.0f)*(float)N*(p[IX3(i+1,j,k)]-p[IX3(i-1,j,k)]);
		v[IX3(i,j,k)] -= (1.0f/2.0f)*(float)N*(p[IX3(i,j+1,k)]-p[IX3(i,j-1,k)]);
		w[IX3(i,j,k)] -= (1.0f/2.0f)*(float)N*(p[IX3(i,j,k+1)]-p[IX3(i,j,k-1)]);
	END_FOR3
	set_bnd3 ( N, 1, u ); set_bnd3 ( N, 2, v ); set_bnd3 ( N, 3, w );
}


void dens_step3 ( int N, float * x, float * x0, float * u, float * v, float * w, float diff, float dt )
{
	add_source3 ( N, x, x0, dt );
	SWAP ( x0, x ); diffuse3 ( N, 0, x, x0, diff, dt );
	SWAP ( x0, x ); advect3 ( N, 0, x, x0, u, v, w, dt );
}


void vel_step3 ( int N, float * u, float * v, float * w, float * u0, float * v0, float * w0, float visc, float dt )
{
	add_source3 ( N, u, u0, dt ); add_source3 ( N, v, v0, dt ); add_source3 ( N, w, w0, dt );
	SWAP ( u0, u ); diffuse3 ( N, 1, u, u0, visc, dt );
	SWAP ( v0, v ); diffuse3 ( N, 2, v, v0, visc, dt );
	SWAP ( w0, w ); diffuse3 ( N, 3, w, w0, visc, dt );	
	project3 ( N, u, v, w, u0, v0 );
	SWAP ( u0, u ); SWAP ( v0, v ); SWAP ( w0, w );
	advect3 ( N, 1, u, u0, u0, v0, w0, dt ); advect3 ( N, 2, v, v0, u0, v0, w0, dt ); advect3 ( N, 3, w, w0, u0, v0, w0, dt );
	project3 ( N, u, v, w, u0, v0 );
}



static void free_data ( void )
{
	if ( u ) free ( u );
	if ( v ) free ( v );
	if ( w ) free ( w );
	if ( u_prev ) free ( u_prev );
	if ( v_prev ) free ( v_prev );
	if ( w_prev ) free ( w_prev );
	if ( dens ) free ( dens );
	if ( dens_prev ) free ( dens_prev );
	if ( object ) free ( object );
}


void make_object()
{
	int i, size=(N+2)*(N+2)*(N+2);
	int x,y,z;

	obj_s = N/5;
	obj_x = N*0.3;
	obj_y = N/2;
	obj_z = N/2;

	for ( i=0 ; i<size ; i++ ) {
		object[i] = 0;
	}
	
	for ( x = obj_x; x<obj_x+obj_s; x ++)
	{
		for ( y = obj_y; y<obj_y+obj_s; y++)
		{
			for ( z = obj_z; z<obj_z+obj_s; z++)
			{
				object[IX3(x,y,z)] = 1;
			}
		}
	}	
}


static void clear_data3 ( void )
{
	int i, size=(N+2)*(N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = w[i] = u_prev[i] = v_prev[i] = w_prev[i] = dens[i] = dens_prev[i] = 0.0f;
	}
	make_object();
}



static int allocate_data3 ( void )
{
	int size = (N+2)*(N+2)*(N+2);

	u			= (float *) malloc ( size*sizeof(float) );
	v			= (float *) malloc ( size*sizeof(float) );
	w			= (float *) malloc ( size*sizeof(float) );
	u_prev		= (float *) malloc ( size*sizeof(float) );
	v_prev		= (float *) malloc ( size*sizeof(float) );
	w_prev		= (float *) malloc ( size*sizeof(float) );
	dens		= (float *) malloc ( size*sizeof(float) );	
	dens_prev	= (float *) malloc ( size*sizeof(float) );
	object  	= (int *)   malloc ( size*sizeof(int)   );

	if ( !u || !v || !w || !u_prev || !v_prev || !w_prev || !dens || !dens_prev ) {
		fprintf ( stderr, "cannot allocate data\n" );
		return ( 0 );
	}

	return ( 1 );
}






static void draw_velocity3 ( void )
{
	int i, j, k;
	float x, y, h, z;	

	h = 1.0f/N;

	glColor4f ( 1.0f, 0.0f, 0.0f, 0.4f );
	glLineWidth ( 1.0f );

	glBegin ( GL_LINES );

		for ( i=1 ; i<=N ; i++ ) {
			x = (i-0.5f)*h;
			for ( j=1 ; j<=N ; j++ ) {
				y = (j-0.5f)*h;
				for ( k=0 ; k<=N ; k++ ) {
					z = (k-0.5f)*h;

					glVertex3f ( x, y, z );
					//glVertex3f ( x+u[IX3(i,j,k)], y+v[IX3(i,j,k)], -10*z+0.1f );
					glVertex3f ( x+u[IX3(i,j,k)], y+v[IX3(i,j,k)], z+w[IX3(i,j,k)] );
				}
			}
		}

	glEnd ();
}


static void draw_density3 ( int art )
{

	int i, j, k;
	float x, y, z, h, h2, d000,d001,d010,d011,d100,d101,d110,d111;
	float col;

	h = 1.0f/N;
	h2 = h * 0.7f;

	glBegin ( GL_QUADS );

		for ( i=0 ; i<=N ; i++ ) {
			x = (i-0.5f)*h;
			for ( j=0 ; j<=N ; j++ ) {
				y = (j-0.5f)*h;
				for ( k=0 ; k<=(N+1) ; k++ ) {
					z = (k-0.5f)*h;
				

					d000 = dens[IX3(i  ,j  ,k  )] * 2 / N;
					d001 = dens[IX3(i  ,j  ,k+1)] * 2 / N;
					d010 = dens[IX3(i  ,j+1,k  )] * 2 / N;
					d011 = dens[IX3(i  ,j+1,k+1)] * 2 / N;
					d100 = dens[IX3(i+1,j  ,k  )] * 2 / N;
					d101 = dens[IX3(i+1,j  ,k+1)] * 2 / N;
					d110 = dens[IX3(i+1,j+1,k  )] * 2 / N;
					d111 = dens[IX3(i+1,j+1,k+1)] * 2 / N;

					col = 1.0f;

					if ( object[IX3(i,j,k)] > 0 ) {
						col = 0.5;
						//d000 = d001 = d010 = d011 = d100 = d101 = d110 = d111 = 1;
						d000 = 1;
					}

			

					if ( art == 0 )
					{

						glColor4f(col, col, 1.0f, d110); glVertex3f(x+ h2/2,y+ h2/2,z+-h2/2);			// Top Righ2t Of Th2e Quad (Top)
						glColor4f(col, col, 1.0f, d010); glVertex3f(x+-h2/2,y+ h2/2,z+-h2/2);			// Top Left Of Th2e Quad (Top)
						glColor4f(col, col, 1.0f, d011); glVertex3f(x+-h2/2,y+ h2/2,z+ h2/2);			// Bottom Left Of Th2e Quad (Top)
						glColor4f(col, col, 1.0f, d111); glVertex3f(x+ h2/2,y+ h2/2,z+ h2/2);			// Bottom Righ2t Of Th2e Quad (Top)

						glColor4f(col, col, 1.0f, d101); glVertex3f(x+ h2/2,y+-h2/2,z+ h2/2);			// Top Righ2t Of Th2e Quad (Bottom)
						glColor4f(col, col, 1.0f, d001); glVertex3f(x+-h2/2,y+-h2/2,z+ h2/2);			// Top Left Of Th2e Quad (Bottom)
						glColor4f(col, col, 1.0f, d000); glVertex3f(x+-h2/2,y+-h2/2,z+-h2/2);			// Bottom Left Of Th2e Quad (Bottom)
						glColor4f(col, col, 1.0f, d100); glVertex3f(x+ h2/2,y+-h2/2,z+-h2/2);			// Bottom Righ2t Of Th2e Quad (Bottom)

						glColor4f(col, col, 1.0f, d111); glVertex3f(x+ h2/2,y+ h2/2,z+ h2/2);			// Top Righ2t Of Th2e Quad (Front)
						glColor4f(col, col, 1.0f, d011); glVertex3f(x+-h2/2,y+ h2/2,z+ h2/2);			// Top Left Of Th2e Quad (Front)
						glColor4f(col, col, 1.0f, d001); glVertex3f(x+-h2/2,y+-h2/2,z+ h2/2);			// Bottom Left Of Th2e Quad (Front)
						glColor4f(col, col, 1.0f, d101); glVertex3f(x+ h2/2,y+-h2/2,z+ h2/2);			// Bottom Righ2t Of Th2e Quad (Front)

						glColor4f(col, col, 1.0f, d100); glVertex3f(x+ h2/2,y+-h2/2,z+-h2/2);			// Bottom Left Of Th2e Quad (Back)
						glColor4f(col, col, 1.0f, d000); glVertex3f(x+-h2/2,y+-h2/2,z+-h2/2);			// Bottom Righ2t Of Th2e Quad (Back)
						glColor4f(col, col, 1.0f, d010); glVertex3f(x+-h2/2,y+ h2/2,z+-h2/2);			// Top Righ2t Of Th2e Quad (Back)
						glColor4f(col, col, 1.0f, d110); glVertex3f(x+ h2/2,y+ h2/2,z+-h2/2);			// Top Left Of Th2e Quad (Back)

						glColor4f(col, col, 1.0f, d011); glVertex3f(x+-h2/2,y+ h2/2,z+ h2/2);			// Top Righ2t Of Th2e Quad (Left)
						glColor4f(col, col, 1.0f, d010); glVertex3f(x+-h2/2,y+ h2/2,z+-h2/2);			// Top Left Of Th2e Quad (Left)
						glColor4f(col, col, 1.0f, d000); glVertex3f(x+-h2/2,y+-h2/2,z+-h2/2);			// Bottom Left Of Th2e Quad (Left)
						glColor4f(col, col, 1.0f, d001); glVertex3f(x+-h2/2,y+-h2/2,z+ h2/2);			// Bottom Righ2t Of Th2e Quad (Left)

						glColor4f(col, col, 1.0f, d110); glVertex3f(x+ h2/2,y+ h2/2,z+-h2/2);			// Top Righ2t Of Th2e Quad (Righ2t)
						glColor4f(col, col, 1.0f, d111); glVertex3f(x+ h2/2,y+ h2/2,z+ h2/2);			// Top Left Of Th2e Quad (Righ2t)
						glColor4f(col, col, 1.0f, d101); glVertex3f(x+ h2/2,y+-h2/2,z+ h2/2);			// Bottom Left Of Th2e Quad (Righ2t)
						glColor4f(col, col, 1.0f, d100); glVertex3f(x+ h2/2,y+-h2/2,z+-h2/2);			// Bottom Righ2t Of Th2e Quad (Righ2t

					}

					if ( art == 1 )
					{

						glColor4f(col, col, 1.0f, d000); glVertex3f(x+-h/2,y+-h/2,z+-h/2);
						glColor4f(col, col, 1.0f, d001); glVertex3f(x+-h/2,y+-h/2,z+ h/2);
						glColor4f(col, col, 1.0f, d011); glVertex3f(x+-h/2,y+ h/2,z+ h/2);
						glColor4f(col, col, 1.0f, d010); glVertex3f(x+-h/2,y+ h/2,z+-h/2);

						glColor4f(col, col, 1.0f, d000); glVertex3f(x+-h/2,y+-h/2,z+-h/2);
						glColor4f(col, col, 1.0f, d100); glVertex3f(x+ h/2,y+-h/2,z+-h/2);
						glColor4f(col, col, 1.0f, d110); glVertex3f(x+ h/2,y+ h/2,z+-h/2);
						glColor4f(col, col, 1.0f, d010); glVertex3f(x+-h/2,y+ h/2,z+-h/2);

						glColor4f(col, col, 1.0f, d000); glVertex3f(x+-h/2,y+-h/2,z+-h/2);
						glColor4f(col, col, 1.0f, d001); glVertex3f(x+-h/2,y+-h/2,z+ h/2);
						glColor4f(col, col, 1.0f, d101); glVertex3f(x+ h/2,y+-h/2,z+ h/2);
						glColor4f(col, col, 1.0f, d100); glVertex3f(x+ h/2,y+-h/2,z+-h/2);

					}

					if ( art == 2 )
					{
		
						d000 = 0.05 + dens[IX3(i,j,k)];
						d010 = dens[IX3(i,j+1,k)];
						d100 = dens[IX3(i+1,j,k)];
						d110 = dens[IX3(i+1,j+1,k)];
						d000 /= 4; d010 /= 4; d100 /= 4; d110 /= 4;

						glColor4f ( col, col, 1.0f, d000 ); glVertex3f ( x,   y,  z_stretch*z );
						glColor4f ( col, col, 1.0f, d100 ); glVertex3f ( x+h, y,  z_stretch*z );
						glColor4f ( col, col, 1.0f, d110 ); glVertex3f ( x+h, y+h,z_stretch*z );
						glColor4f ( col, col, 1.0f, d010 ); glVertex3f ( x,   y+h,z_stretch*z );

					}			
				}
			}
		}

	glEnd ();
}




float zufallszahl( float ende )
{
	float wert;
	wert = ( rand() * ende ) / 32767;
	return wert;
}

static void mache_durcheinander ( float * d, float * u, float * v, float staerke )
{
	int i, j, size = (N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] = 0.0f;
	}


	i = (int)zufallszahl(N-1)+1;
	j = (int)zufallszahl(N-1+1);

	if ( i<1 || i>N || j<1 || j>N ) return;

	
		u[IX(i,j)] = staerke * force * (zufallszahl(10)-5);
		v[IX(i,j)] = staerke * force * (zufallszahl(10)-5);
		
		d[IX(i,j)] = source * staerke;
		


	return;
}


static void schornstein3 ( float * d, float * u, float * v, float * w, float staerke )
{
	int i, j, size = (N+2)*(N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = d[i] = 0.0f;
	}

	i = N/2;
	j = N/4;

	if ( i<1 || i>N || j<1 || j>N ) return;

	



	u[IX3(i,j,N/2)] = 0;
	v[IX3(i,j,N/2)] = 0;	
	//v[IX3(i,j,N/2)] = staerke * force * 2;
	w[IX3(i,j,N/2)] = 0;
	//u[IX3(i,j,N/2)] = staerke * force * 1.07;
	v[IX3(i,j,N/2)] = staerke * force * 0.9;
	//w[IX3(i,j,N/2)] = 0.4;
	d[IX3(i,j,N/2)] = source * staerke / 15;

	
	return;
}



static void mache_ordnung3 ( float * d, float * u, float * v, float * w )
{
	int i, size = (N+2)*(N+2)*(N+2);

	for ( i=0 ; i<size ; i++ ) {
		u[i] = v[i] = w[i] = d[i] = 0.0f;
	}
	return;
}


void output(float x, float y, char *string)
{
	int len, i;
	glRasterPos3f(x, y, -1);
	len = (int) strlen(string);
	for (i = 0; i < len; i++) {
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, string[i]);
	}
}




void ventilator_density()
{
	int y,z;
	//for ( i=1 ; i<=N ; i++ ) {
		//for ( j=1 ; j<=N ; j++ ) {			
	for ( y = obj_y; y<obj_y+obj_s; y++)
	{
		for ( z = obj_z; z<obj_z+obj_s; z++)
		{
			dens_prev[IX3(1  ,y,z)] = 0.1f;
		}
	}	
}


void renderScene(void)
{
	char str[255];
	time_t seconds;

	seconds = time (NULL);

	sekunden = seconds;
	framecounter ++;
	if ( sekunden > sekunden_alt ) {
		fps = framecounter;
		framecounter = 0;
	}
	sekunden_alt = sekunden;
  
	if (mouseActive==1)
	{
		POINT val = mouse.getRelativeMovement();
		cameraControl.mouseTurn(val.x);
		cameraControl.mouseTilt(-val.y);
	}

	cameraControl.checkCurrentMovement();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
		glLoadIdentity();
		glColor3f(0.6f, 0.7f, 1.0f);
		sprintf (str,"n=%i; fps=%i", N, fps);
		output(-0.49f,-0.4f,str);
	glPopMatrix();
	

	
	if ( dvel ) draw_velocity3 ();
	else {

		draw_density3(draw_dens_mode);
	}
		

	//if (schornstein_an)  schornstein3(dens_prev, u_prev, v_prev, w_prev, 1);
	ventilator_density();
	vel_step3 ( N, u, v, w, u_prev, v_prev, w_prev, visc, dt );
	dens_step3 ( N, dens, dens_prev, u, v, w, diff, dt );
	mache_ordnung3 ( dens_prev, u_prev, v_prev, w_prev );


	glutSwapBuffers();
}






void main(int argc, char *argv[])
{
	

	dt = 0.1f;
	diff = 0.0f;
	visc = 0.0f;
	force = 5.0f;
	source = 100.0f;

	if ( !allocate_data3 () )	exit ( 1 );
	clear_data3 ();


	srand( (unsigned)time( NULL ) );


	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowPosition(100, 100);
	glutInitWindowSize(win_width, win_height);	
	glutCreateWindow("Tobias Hermann");


	cameraControl.eyePosZ = 2.4;
	cameraControl.eyePosX = 0.5;
	cameraControl.eyePosY = 0.5;

	glClearColor(0.0f, 0.0f, 0.0f, 0.5f);				// Black Background
	glClearDepth(1.0f);									// Depth Buffer Setup

		
	initEnvironment();
	glutMainLoop();

	free_data ();
}

void initEnvironment(void)
{
	glutDisplayFunc(renderScene);
	glutIdleFunc(renderScene);
	glutReshapeFunc(changePerspective);
	glutIgnoreKeyRepeat(1);
	glutKeyboardFunc(normalKeyPressed);
	glutKeyboardUpFunc(normalKeyReleased);
	glutSpecialFunc(specialKeyPressed);
	glutSpecialUpFunc(specialKeyReleased);
	glEnable(GL_BLEND);			// Turn Blending On
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);
	glDisable(GL_DEPTH_TEST);	// Turn Depth Testing Off
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	
}



void changePerspective(int width, int height)
{
	if (height == 0) height = 1;
	float ratio = 1.0f * width / height;
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0, 0, width, height);
	gluPerspective(45.0f, ratio, 0.1f, 1000.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	cameraControl.updateCamera();
}


void normalKeyPressed(unsigned char key, int x, int y)
{
	switch (key)
	{
		//Programm mit ESC beenden
		case 27:
			free_data ();
			exit(0);
			break;
		// KAMERA Steuerung
		case 'w':
			//cameraControl.walkForward();
			cameraControl.flyForward();
			break;
		
		case 's':
			//cameraControl.walkBackward();
			cameraControl.flyBackward();
			break;
			
		case 'a':
			cameraControl.strafeLeft();
			break;
		
		case 'd':
			cameraControl.strafeRight();
			break;

		case 'e':
			einatmen ++;			
			break;

		case 'h':
			draw_dens_mode ++;
			draw_dens_mode = draw_dens_mode % 3;

		case 'r':
			mache_durcheinander ( dens_prev, u_prev, v_prev, einatmen );
			einatmen = 0;
			break;

		case 't':
			schornstein_an = !schornstein_an;
			break;


		case 'c':
		case 'C':
			clear_data3 ();
			break;
	
		case 'v':
		case 'V':
			dvel = !dvel;
			break;
		case '+':
			N++;
			free_data();
			allocate_data3();
			clear_data3();
			break;
		case '-':
			if ( N > 10 ) {
				N--;
				free_data();
				allocate_data3();
				clear_data3();
			}
			break;
	}
}

void normalKeyReleased(unsigned char key, int x, int y)
{
	switch (key)
	{
		// KAMERA Steuerung
		case 'w':
		case 'W':
		case 's':
		case 'S':
			cameraControl.stopWalking();
			cameraControl.stopFlying();
			break;
		
		case 'a':
		case 'A':
		case 'd':
		case 'D':
			cameraControl.stopStrafing();
			break;	
		case 'g':
			if (isGridActive)
				isGridActive = 0;
			else
				isGridActive = 1;
			break;

		case ' ':
			if (mouseActive == 1)
			{
				glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
				mouseActive = 0;
			}
			else
			{
				glutSetCursor(GLUT_CURSOR_NONE);
				mouseActive = 1;
			}
			break;	
	}
}

void specialKeyPressed(int key, int x, int y)
{
	int modifiers = glutGetModifiers();

	switch (key)
	{
		// Zwischen Fullscreen und Window wechseln
		case GLUT_KEY_F11:
			if (glutGameModeGet(GLUT_GAME_MODE_ACTIVE))
			{
				glutLeaveGameMode();
			}
			else
			{
				glutGameModeString(GAMEMODE_STRING);
				if (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE))
				{
					glutEnterGameMode();

					if (mouseActive == 1)
						glutSetCursor(GLUT_CURSOR_NONE);
					else
						glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
										
					mouse.resetMouse();
					initEnvironment();
				}
				else
					exit(1);
			}
			break;
		
		// KAMERA Steuerung
		case GLUT_KEY_LEFT:
			cameraControl.turnLeft();
			break;
		
		case GLUT_KEY_RIGHT:
			cameraControl.turnRight();
			break;
		
		case GLUT_KEY_UP:
			cameraControl.tiltUp();
			break;
		
		case GLUT_KEY_DOWN:
			cameraControl.tiltDown();
			break;

		case GLUT_KEY_PAGE_UP:
			cameraControl.hoverUp();
			break;
		
		case GLUT_KEY_PAGE_DOWN:
			cameraControl.hoverDown();
			break;
	}
}

void specialKeyReleased(int key, int x, int y)
{
	switch (key)
	{
		// KAMERA Steuerung
		case GLUT_KEY_LEFT:
		case GLUT_KEY_RIGHT:
			cameraControl.stopTurning();
			break;
		
		case GLUT_KEY_UP:
		case GLUT_KEY_DOWN:
			cameraControl.stopTilting();
			break;
		
		case GLUT_KEY_PAGE_UP:
		case GLUT_KEY_PAGE_DOWN:
			cameraControl.stopHovering();
			break;
	}
}